import torch
from torch import profiler

from .._utils import _validate_parameters
from ..potentials import Potential


class Calculator(torch.nn.Module):
    """
    Base calculator for the torch interface. Based on a
    :class:`Potential` class, it computes the value of a potential
    by either directly summing over neighbor atoms, or by combining
    a local part computed in real space, and a long-range part computed
    in the Fourier domain. The class can be used directly to evaluate
    the real-space part of the potential, or subclassed providing
    a strategy to evalate the long-range contribution in k-space
    (see e.g. :class:`PMECalculator` or :class:`EwaldCalculator`).
    NB: typically a subclass should only provide an implementation of
    :func:`Calculator._compute_kspace`.

    :param potential: a :class:`Potential` class object containing the functions
        that are necessary to compute the various components of the potential, as
        well as the parameters that determine the behavior of the potential itself.
    :param full_neighbor_list: parameter indicating whether the neighbor information
        will come from a full (True) or half (False, default) neighbor list.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.
    """

    def __init__(
        self,
        potential: Potential,
        full_neighbor_list: bool = False,
        prefactor: float = 1.0,
    ):
        super().__init__()

        if not isinstance(potential, Potential):
            raise TypeError(
                f"Potential must be an instance of Potential, got {type(potential)}"
            )

        self.potential = potential
        self.full_neighbor_list = full_neighbor_list
        self.prefactor = prefactor

    def _compute_rspace(
        self,
        charges: torch.Tensor,
        neighbor_indices: torch.Tensor,
        neighbor_distances: torch.Tensor,
    ) -> torch.Tensor:
        # Compute the pair potential terms V(r_ij) for each pair of atoms (i,j)
        # contained in the neighbor list
        with profiler.record_function("compute bare potential"):
            if self.potential.smearing is None:
                potentials_bare = self.potential.from_dist(neighbor_distances)
            else:
                potentials_bare = self.potential.sr_from_dist(neighbor_distances)

        # Multiply the bare potential terms V(r_ij) with the corresponding charges
        # of ``atom j'' to obtain q_j*V(r_ij). Since each atom j can be a neighbor of
        # multiple atom i's, we need to access those from neighbor_indices
        atom_is = neighbor_indices[:, 0]
        atom_js = neighbor_indices[:, 1]
        with profiler.record_function("compute real potential"):
            contributions_is = charges[atom_js] * potentials_bare.unsqueeze(-1)

        # For each atom i, add up all contributions of the form q_j*V(r_ij) for j
        # ranging over all of its neighbors.
        with profiler.record_function("assign potential"):
            potential = torch.zeros_like(charges)
            potential.index_add_(0, atom_is, contributions_is)
            # If we are using a half neighbor list, we need to add the contributions
            # from the "inverse" pairs (j, i) to the atoms i
            if not self.full_neighbor_list:
                contributions_js = charges[atom_is] * potentials_bare.unsqueeze(-1)
                potential.index_add_(0, atom_js, contributions_js)

        # Compensate for double counting of pairs (i,j) and (j,i)
        return potential / 2

    def _compute_kspace(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
    ) -> torch.Tensor:
        """
        Computes the Fourier-domain contribution to the potential, typically
        corresponding to a long-range, slowly decaying type of interaction.
        """
        raise NotImplementedError(
            f"`compute_kspace` not implemented for {self.__class__.__name__}"
        )

    def forward(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
        neighbor_indices: torch.Tensor,
        neighbor_distances: torch.Tensor,
    ):
        r"""
        Compute the potential "energy".

        It is calculated as

        .. math::

            V_i = \frac{1}{2} \sum_{j} q_j\,v(r_{ij})

        where :math:`v(r)` is the pair potential defined by the ``potential`` parameter
        and :math:`q_j` are atomic "charges" (corresponding to the electrostatic charge
        when using a Coulomb potential).

        If the ``smearing`` of the ``potential`` is not set, the calculator evaluates
        only the real-space part of the potential. Otherwise, provided that the
        calculator implements a ``_compute_kspace`` method, it will also evaluate the
        long-range part using a Fourier-domain method.

        :param charges: torch.tensor of shape ``(n_channels, len(positions))``
            containaing the atomic (pseudo-)charges. ``n_channels`` is the number of
            charge channels the potential should be calculated. For a standard potential
            ``n_channels = 1``. If more than one "channel" is provided multiple
            potentials for the same position are computed depending on the charges and
            the potentials.
        :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th
            basis vector of the unit cell
        :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
            coordinates of the ``N`` particles within the supercell.
        :param neighbor_indices: torch.tensor with the ``i,j`` indices of neighbors for
            which the potential should be computed in real space.
        :param neighbor_distances: torch.tensor with the pair distances of the neighbors
            for which the potential should be computed in real space.
        """
        _validate_parameters(
            charges=charges,
            cell=cell,
            positions=positions,
            neighbor_indices=neighbor_indices,
            neighbor_distances=neighbor_distances,
            smearing=self.potential.smearing,
        )

        # Compute short-range (SR) part using a real space sum
        potential_sr = self._compute_rspace(
            charges=charges,
            neighbor_indices=neighbor_indices,
            neighbor_distances=neighbor_distances,
        )

        if self.potential.smearing is None:
            return self.prefactor * potential_sr
        # Compute long-range (LR) part using a Fourier / reciprocal space sum
        potential_lr = self._compute_kspace(
            charges=charges,
            cell=cell,
            positions=positions,
        )

        return self.prefactor * (potential_sr + potential_lr)



from typing import Optional

import torch

from ..lib import (
    CubicSpline,
    CubicSplineReciprocal,
    compute_second_derivatives,
    compute_spline_ft,
)
from .potential import Potential


class SplinePotential(Potential):
    r"""
    Potential built from a spline interpolation.

    The potential is assumed to have only a long-range part, but one can also
    add a short-range part if needed, by inheriting and redefining
    ``sr_from_dist``.
    The real-space potential is computed based on a cubic spline built at
    initialization time. The Fourier-domain kernel is computed numerically
    as a spline, too.  Assumes the infinite-separation value of the
    potential to be zero.

    :param r_grid: radial grid for the real-space evaluation
    :param y_grid: potential values for the real-space evaluation
    :param k_grid: radial grid for the k-space evaluation;
        computed automatically from ``r_grid`` if absent.
    :param yhat_grid: potential values for the k-space evaluation;
        computed automatically from ``y_grid`` if absent.
    :param reciprocal: flag that determines if the splining should
        be performed on a :math:`1/r` axis; suitable to describe
        long-range potentials. ``r_grid`` should contain only
        stricty positive values.
    :param y_at_zero: value to be used for :math:`r\rightarrow 0`
        when using a reciprocal spline
    :param yhat_at_zero: value to be used for :math:`k\rightarrow 0`
        in the k-space kernel
    :param smearing: The length scale associated with the switching between
        :math:`V_{\mathrm{SR}}(r)` and :math:`V_{\mathrm{LR}}(r)`
    :param exclusion_radius: A length scale that defines a *local environment* within
        which the potential should be smoothly zeroed out, as it will be described by a
        separate model.
    :param exclusion_degree: Controls the sharpness of the transition in the cutoff function
        applied within the ``exclusion_radius``. The cutoff is computed as a raised cosine
        with exponent ``exclusion_degree``
    """

    def __init__(
        self,
        r_grid: torch.Tensor,
        y_grid: torch.Tensor,
        k_grid: Optional[torch.Tensor] = None,
        yhat_grid: Optional[torch.Tensor] = None,
        reciprocal: Optional[bool] = False,
        y_at_zero: Optional[float] = None,
        yhat_at_zero: Optional[float] = None,
        smearing: Optional[float] = None,
        exclusion_radius: Optional[float] = None,
        exclusion_degree: int = 1,
    ):
        super().__init__(
            smearing=smearing,
            exclusion_radius=exclusion_radius,
            exclusion_degree=exclusion_degree,
        )

        if len(y_grid) != len(r_grid):
            raise ValueError("Length of radial grid and value array mismatch.")

        self.register_buffer("r_grid", r_grid)
        self.register_buffer("y_grid", y_grid)

        if reciprocal:
            if torch.min(r_grid) <= 0.0:
                raise ValueError(
                    "Positive-valued radial grid is needed for reciprocal axis spline."
                )
            self._spline = CubicSplineReciprocal(r_grid, y_grid, y_at_zero=y_at_zero)
        else:
            self._spline = CubicSpline(r_grid, y_grid)

        if k_grid is None:
            # defaults to 2pi/r_grid_points if reciprocal, to r_grid if not
            if reciprocal:
                k_grid = torch.pi * 2 * torch.reciprocal(r_grid).flip(dims=[0])
            else:
                k_grid = r_grid.clone().detach()

        self.register_buffer("k_grid", k_grid)

        if yhat_grid is None:
            # computes automatically!
            yhat_grid = compute_spline_ft(
                k_grid,
                r_grid,
                y_grid,
                compute_second_derivatives(r_grid, y_grid),
            )

        self.register_buffer("yhat_grid", yhat_grid)

        # the function is defined for k**2, so we define the grid accordingly
        if reciprocal:
            self._krn_spline = CubicSplineReciprocal(
                k_grid**2, yhat_grid, y_at_zero=yhat_at_zero
            )
        else:
            self._krn_spline = CubicSpline(k_grid**2, yhat_grid)

        if y_at_zero is None:
            self._y_at_zero = self._spline(
                torch.zeros(1, dtype=self.r_grid.dtype, device=self.r_grid.device)
            )
        else:
            self._y_at_zero = torch.tensor(
                y_at_zero, dtype=self.r_grid.dtype, device=self.r_grid.device
            )

        if yhat_at_zero is None:
            self._yhat_at_zero = self._krn_spline(
                torch.zeros(1, dtype=self.k_grid.dtype, device=self.k_grid.device)
            )
        else:
            self._yhat_at_zero = torch.tensor(
                yhat_at_zero, dtype=self.k_grid.dtype, device=self.k_grid.device
            )

    def from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        # if the full spline is not given, falls back on the lr part
        return self.lr_from_dist(dist) + self.sr_from_dist(dist)

    def sr_from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        """
        Short-range part of the range-separated potential.

        :param dist: torch.tensor containing the distances at which the potential is to
            be evaluated.
        """
        return 0.0 * dist

    def lr_from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        return self._spline(dist)

    def lr_from_k_sq(self, k_sq: torch.Tensor) -> torch.Tensor:
        return self._krn_spline(k_sq)

    def self_contribution(self) -> torch.Tensor:
        return self._y_at_zero

    def background_correction(self) -> torch.Tensor:
        return torch.zeros(1)

    from_dist.__doc__ = Potential.from_dist.__doc__
    lr_from_dist.__doc__ = Potential.lr_from_dist.__doc__
    lr_from_k_sq.__doc__ = Potential.lr_from_k_sq.__doc__
    self_contribution.__doc__ = Potential.self_contribution.__doc__
    background_correction.__doc__ = Potential.background_correction.__doc__



import torch

try:
    from metatensor.torch import Labels, TensorBlock, TensorMap
except ImportError:
    raise ImportError(
        "metatensor.torch is required for torchpme.metatensor but is not installed. "
        "Try installing it with:\npip install metatensor[torch]"
    ) from None

try:
    from metatomic.torch import System
except ImportError:
    raise ImportError(
        "metatomic is required for torchpme.metatensor but is not installed. "
        "Try installing it with:\npip install metatomic"
    ) from None

from .. import calculators as torch_calculators


class Calculator(torch.nn.Module):
    """
    Base calculator for the metatensor interface.

    This is just a thin wrapper around the corresponding generic torch
    :class:`torchpme.calculators.Calculator`. If you want to wrap a ``metatensor``
    interface around another calculator, you can just define the class and set the
    static member ``_base_calculator`` to the corresponding torch calculator.
    """

    _base_calculator: type[torch_calculators.Calculator] = torch_calculators.Calculator

    def __init__(self, *args, **kwargs):
        super().__init__()

        self._calculator = self._base_calculator(*args, **kwargs)

    @staticmethod
    def _validate_compute_parameters(system: System, neighbors: TensorBlock) -> None:
        dtype = system.positions.dtype
        device = system.positions.device

        if neighbors.values.dtype != dtype:
            raise ValueError(
                f"dtype of `neighbors` ({neighbors.values.dtype}) must be the same "
                f"as `system` ({dtype})"
            )

        if neighbors.values.device != device:
            raise ValueError(
                f"device of `neighbors` ({neighbors.values.device}) must be the same "
                f"as `system` ({device})"
            )

        # Check metadata of neighbors
        samples_names = neighbors.samples.names
        if (
            len(samples_names) != 5
            or samples_names[0] != "first_atom"
            or samples_names[1] != "second_atom"
            or samples_names[2] != "cell_shift_a"
            or samples_names[3] != "cell_shift_b"
            or samples_names[4] != "cell_shift_c"
        ):
            raise ValueError(
                "Invalid samples for `neighbors`: the sample names must be "
                "'first_atom', 'second_atom', 'cell_shift_a', 'cell_shift_b', "
                "'cell_shift_c'"
            )

        components_labels = Labels(
            ["xyz"],
            torch.arange(3, dtype=torch.int32, device=device).unsqueeze(1),
        )
        components = neighbors.components
        if len(components) != 1 or components[0] != components_labels:
            raise ValueError(
                "Invalid components for `neighbors`: there should be a single "
                "'xyz'=[0, 1, 2] component"
            )

        properties_labels = Labels(
            ["distance"], torch.zeros(1, 1, dtype=torch.int32, device=device)
        )
        if neighbors.properties != properties_labels:
            raise ValueError(
                "Invalid properties for `neighbors`: there should be a single "
                "'distance'=0 property"
            )

        if "charges" not in system.known_data():
            raise ValueError("`system` does not contain `charges` data")

        charge_tensor = system.get_data("charges")
        if len(charge_tensor) != 1:
            raise ValueError(
                f"Charge tensor have exactlty one block but has {len(charge_tensor)} "
                "blocks"
            )

        n_charge_components = len(charge_tensor.block().components)
        if n_charge_components > 0:
            raise ValueError(
                "TensorBlock containg the charges should not have components; "
                f"found {n_charge_components}"
            )

    def forward(self, system: System, neighbors: TensorBlock) -> TensorMap:
        """
        Compute the potential "energy".

        The ``system`` must contain a custom data field ``charges``. The potential will
        be calculated for each ``"charges_channel"``, which will also be the properties
        name of the returned :class:`metatensor.torch.TensorMap`.

        :param system: System to run the calculations. The system must have attached
            ``"charges"`` using the :meth:`add_data
            <metatomic.torch.System.add_data>` method.
        :param neighbors: The neighbor list. If a neighbor list is attached to a
            :class:`metatomic.torch.System` it can be extracted with the
            :meth:`get_neighborlist
            <metatomic.torch.System.get_neighborlist>` method using a
            :class:`NeighborListOptions
            <metatomic.torch.NeighborListOptions>`. Note to use the same
            ``full_list`` option for these options as provided for
            ``full_neighbor_list`` in the constructor.

            .. note::

                Although ``neighbors`` can be attached to the ``system``, they are
                required to be passed explicitly here. While it's possible to design the
                class to automatically extract the neighbor list by accepting a
                :class:`NeighborListOptions
                <metatomic.torch.NeighborListOptions>` directly in the
                constructor, we chose explicit passing for consistency with the torch
                interface.

        :return: :class:`metatensor.torch.TensorMap` containing the potential
        """
        self._validate_compute_parameters(system, neighbors)

        device = system.positions.device
        charges = system.get_data("charges").block().values

        n_atoms = len(system)
        samples = torch.zeros((n_atoms, 2), device=device, dtype=torch.int32)
        samples[:, 0] = 0
        samples[:, 1] = torch.arange(n_atoms, device=device, dtype=torch.int32)

        neighbor_indices = neighbors.samples.view(["first_atom", "second_atom"]).values

        if device.type == "cpu":
            # move to 64-bit integers, for some reason indexing 64-bit is a lot faster
            # than using 32-bit integers on CPU. CUDA seems fine with either types
            neighbor_indices = neighbor_indices.to(
                torch.int64, memory_format=torch.contiguous_format
            )

        neighbor_distances = torch.linalg.norm(neighbors.values, dim=1).squeeze(1)

        potential = self._calculator.forward(
            charges=charges,
            cell=system.cell,
            positions=system.positions,
            neighbor_indices=neighbor_indices,
            neighbor_distances=neighbor_distances,
        )

        properties_values = torch.arange(
            charges.shape[1], device=device, dtype=torch.int32
        )

        block = TensorBlock(
            values=potential,
            samples=Labels(["system", "atom"], samples),
            components=[],
            properties=Labels("charges_channel", properties_values.unsqueeze(1)),
        )

        keys = Labels("_", torch.zeros(1, 1, dtype=torch.int32, device=device))
        return TensorMap(keys=keys, blocks=[block])
